@@ -21,16 +21,21 @@
 
 
 static GMainLoop *loop;
+static DBusGConnection *dbusconn;
 
 /* Generate the GObject boilerplate */
 G_DEFINE_TYPE(FMRXEnabler, fmrx_enabler, G_TYPE_OBJECT)
 
 static void daemonize();
+static gboolean is_bluetooth_allowed();
 static gboolean hci_dev_up(int sock, gboolean *had_bluetooth);
 static gboolean hci_dev_down(int sock);
 static gboolean set_i2c_enabled(int sock, gboolean value);
 static gboolean set_rds_enabled(gboolean value);
 static gboolean watchdog_handler(gpointer data);
+static void on_change_device_mode(DBusGProxy *proxy,
+                                  const char *value,
+                                  gpointer data);
 
 
 /* GObject class initializer.
@@ -48,9 +53,9 @@ fmrx_enabler_class_init(FMRXEnablerClass *clss) {
 static void
 fmrx_enabler_init(FMRXEnabler *self) {
 
-    DBusGConnection *conn;
     GError *error = NULL;
     DBusGProxy *driver_proxy;
+    DBusGProxy *mce_proxy;
     guint32 request_name_ret;
     
     self->hci_socket = 0;
@@ -59,15 +64,17 @@ fmrx_enabler_init(FMRXEnabler *self) {
     self->last_request = 0;
 
     /* set up D-Bus service */
-    conn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
-    if (conn == NULL) {
+    dbusconn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
+    if (dbusconn == NULL) {
         g_print("could not connect to D-Bus: %s", error->message);
         g_error_free(error);
         return;
     }
     
-    dbus_g_connection_register_g_object(conn, FMRX_OBJECT_PATH, G_OBJECT(self));
-    driver_proxy = dbus_g_proxy_new_for_name(conn,
+    dbus_g_connection_register_g_object(dbusconn,
+                                        FMRX_OBJECT_PATH,
+                                        G_OBJECT(self));
+    driver_proxy = dbus_g_proxy_new_for_name(dbusconn,
                                              DBUS_SERVICE_DBUS,
                                              DBUS_PATH_DBUS,
                                              DBUS_INTERFACE_DBUS);
@@ -84,9 +91,24 @@ fmrx_enabler_init(FMRXEnabler *self) {
         return;
     }
 
+    /* watch D-Bus for getting notified about device mode changes */
+    /* ... or rather don't as this should be handled by the clients...
+    mce_proxy = dbus_g_proxy_new_for_name(dbusconn,
+                                          MCE_SERVICE,
+                                          MCE_SIGNAL_PATH,
+                                          MCE_SIGNAL_IF);
+    dbus_g_proxy_add_signal(mce_proxy, MCE_DEVICE_MODE_SIG,
+                            G_TYPE_STRING, G_TYPE_INVALID);
+    dbus_g_proxy_connect_signal(mce_proxy, MCE_DEVICE_MODE_SIG,
+                                G_CALLBACK(on_change_device_mode),
+                                (gpointer) self,
+                                NULL);
+    */
+    
 }
 
 
+
 /* Turns this process into a daemon by using the double-(pitch)-fork. :)
  */
 static void
@@ -109,6 +131,48 @@ daemonize() {
 }
 
 
+/* Checks whether Bluetooth is allowed at the moment or not.
+ */
+static gboolean
+is_bluetooth_allowed() {
+
+    DBusGProxy *proxy;
+    GError *error = NULL;
+    char *device_mode;
+    
+    proxy = dbus_g_proxy_new_for_name(dbusconn,
+                                      MCE_SERVICE,
+                                      MCE_REQUEST_PATH,
+                                      MCE_REQUEST_IF);
+    if (! dbus_g_proxy_call(proxy, MCE_DEVICE_MODE_GET, &error,
+                            G_TYPE_INVALID,
+                            G_TYPE_STRING, &device_mode, G_TYPE_INVALID)) {
+
+        g_print("error retrieving device mode\n");
+        g_error_free(error);
+        return FALSE;
+
+    }
+    
+    g_object_unref(proxy);
+
+    /* test for offline or flight mode */
+    if (strcmp(device_mode, MCE_FLIGHT_MODE) == 0 ||
+        strcmp(device_mode, MCE_OFFLINE_MODE) == 0) {
+        
+        g_free(device_mode);
+        return FALSE;
+        
+    } else {
+
+        g_free(device_mode);
+        return TRUE;
+        
+    }
+
+}
+
+
 /* Powers up the Bluetooth device.
  */
 static gboolean
@@ -198,6 +262,29 @@ set_rds_enabled(gboolean value) {
 }
 
 
+/* Signal handler for catching offline mode.
+ */
+static void
+on_change_device_mode(DBusGProxy *proxy,
+                      const char *value,
+                      gpointer data) {
+   
+    FMRXEnabler *self = (FMRXEnabler*) data;
+
+    if (strcmp(value, MCE_FLIGHT_MODE) == 0 ||
+        strcmp(value, MCE_OFFLINE_MODE) == 0) {
+
+        /* we got into offline mode, switch off FM radio */
+        self->driver_loaded = FALSE;
+        close(self->hci_socket);
+        
+        g_print("device entered offline mode\n");
+        
+    }
+
+}
+
+
 /* Watchdog for automatically powering down the radio and I2C when not in use
  * for a while.
  */
@@ -208,6 +295,11 @@ watchdog_handler(gpointer data) {
     GError *error;
     FMRXEnabler *self = (FMRXEnabler*) data;
 
+    /* stop watchdog if driver was unloaded */
+    if (! self->driver_loaded) {
+        return FALSE;
+    }
+
     g_print("watchdog checking\n");
     if (difftime(time(NULL), self->last_request) > UNLOAD_TIMEOUT) {
     
@@ -258,6 +350,15 @@ dbus_fmrx_enabler_request(FMRXEnabler *self,
         *result = RESULT_OK;
         
     } else {
+        /* check if we're allowed to enable Bluetooth */
+        g_print("checking whether Bluetooth may be enabled... ");
+        if (is_bluetooth_allowed() == FALSE) {
+            g_print("no\n");
+            *result = RESULT_BLUETOOTH_DISALLOWED;
+            return TRUE;
+        }
+        g_print("yes\n");
+
         /* attempt to unload module first */
         g_spawn_command_line_sync("/sbin/rmmod radio-bcm2048",
                             NULL, NULL, &status, &error);
